"""Translation logic for structures from YAML
"""

from collections import defaultdict

from packaging.requirements import InvalidRequirement, Requirement
from packaging.version import InvalidVersion, Version

from .abcs import YAMLCandidate
from .exceptions import AggregatedYAMLErrors, MalformedYAML, YAMLException


def translate_specification(test_id, specification):
    """We assume things that have been checked during the expansion.
    """
    index = specification["index"]
    actions = specification["actions"]
    results = specification["results"]

    retval = {}
    retval["candidates"], retval["dependencies"] = _translate_index(test_id, index)
    retval["subtests"] = _translate_actions_and_results(test_id, actions, results)

    return retval


def _translate_list(test_id, argument, name, function):
    retval = []

    errors = AggregatedYAMLErrors()
    for i, item in enumerate(argument):
        try:
            item = function(test_id, i, name, item)
        except YAMLException as e:
            errors.add(e)
        else:
            retval.append(item)

    if errors:
        raise errors

    return retval


def _translate_index(test_id, index):
    first_pass = _translate_list(test_id, index, "index", _translate_index_item)

    # Create a mapping of candidate data
    candidates = defaultdict(list)
    dependencies = {}

    # Error Checking
    errors = AggregatedYAMLErrors()
    for item in first_pass:
        candidate, new_deps = item
        already_specified = set(dependencies) & set(new_deps)
        if already_specified:
            # Nope. There's a problem.
            errors.add(
                MalformedYAML(
                    test_id,
                    "Got double specification for same candidate(s): {}",
                    already_specified,
                )
            )
        else:
            # Add the candidate and dependency information
            candidates[candidate.name].append(candidate)
            dependencies.update(new_deps)

    if errors:
        raise errors

    return dict(candidates), dependencies


def _translate_index_item(test_id, i, _, item):
    name = item["name"]
    try:
        version = Version(item["version"])
    except InvalidVersion:
        raise MalformedYAML(
            test_id, "index[{}] has an invalid version: {}", i, item["version"]
        )

    # Compose Dependencies
    dependencies = {}
    errors = AggregatedYAMLErrors()

    # Top level first.
    key = "{} {}".format(name, version)
    try:
        dependencies[key] = _translate_list(
            test_id, item["depends"], "index[{}], depends".format(i), function=_make_req
        )
    except YAMLException as e:
        errors.add(e)

    # Extras next.
    for extra in item["extras"]:
        key = "{}[{}] {}".format(name, extra, version)
        try:
            dependencies[key] = _translate_list(
                test_id,
                item["extras"][extra],
                "index[{}], extra={!r}".format(i, extra),
                function=_make_req,
            )
        except YAMLException as e:
            errors.add(e)

    if errors:
        raise errors

    return YAMLCandidate(name, version), dependencies


def _make_req(test_id, i, name, req_str):
    try:
        return Requirement(req_str)
    except InvalidRequirement as e:
        raise MalformedYAML(test_id, "{} has invalid requirement {!r}", name, req_str)


def _translate_actions_and_results(test_id, actions, results):
    return list(zip(actions, results))



import os

import pytest

from tests.lib.yaml.fixtures import YamlFixtureFile


def pytest_collect_file(parent, path):
    if path.ext == ".yaml" and ".ignore" not in str(path):
        return YamlFixtureFile(path, parent)


def pytest_collection_modifyitems(session, config, items):
    # Mark Tests based on which directory they live in
    for item in items:
        if isinstance(item, YamlFixtureFile):
            item.add_marker(pytest.mark.yaml)
            item.add_marker(pytest.mark.integration)

        if not hasattr(item, "module"):
            continue

        # Get path of the file
        module_path = os.path.relpath(
            item.module.__file__, os.path.commonprefix([__file__, item.module.__file__])
        )
        # Mark the type of test
        module_root_dir = module_path.split(os.pathsep)[0]
        if module_root_dir.startswith("integration"):
            item.add_marker(pytest.mark.integration)
        elif module_root_dir.startswith("unit"):
            item.add_marker(pytest.mark.unit)
        elif module_root_dir.startswith("lib"):
            item.add_marker(pytest.mark.integration)
        else:
            msg = "Unknown test type (filename = {0}, root_dir = {1})".format(
                module_path, module_root_dir
            )
            raise RuntimeError(msg)



"""Tests for translation of expansions into actual objects for the Provider.
"""

import pytest
from packaging.requirements import Requirement as _Requirement
from packaging.version import Version

from tests.lib.yaml import translate
from tests.lib.yaml.abcs import YAMLCandidate
from tests.lib.yaml.exceptions import MalformedYAML, YAMLException


# XXX: This is a hack for working around the fact that packaging's Requirement doesn't
#      have equality checking.
class Requirement(_Requirement):

    def __eq__(self, other):
        return str(self) == str(other)


def test_translate_specification():
    spec = {
        "index": [
            {
                "name": "A",
                "version": "1.0.0",
                "depends": ["six"],
                "extras": {"docs": ["Sphinx"]},
            }
        ],
        "actions": [{"action": "install", "items": "A"}],
        "results": [{"set": ["A 1.0.0"], "graph": []}],
    }
    expected = {
        "candidates": {"A": [YAMLCandidate("A", Version("1.0.0"))]},
        "dependencies": {
            "A 1.0.0": [Requirement("six")],
            "A[docs] 1.0.0": [Requirement("Sphinx")],
        },
        "subtests": [
            ({"action": "install", "items": "A"}, {"set": ["A 1.0.0"], "graph": []})
        ],
    }
    got = translate.translate_specification(0, spec)

    assert expected == got


def test_index_double_specification():
    with pytest.raises(YAMLException) as e:
        translate._translate_index(
            0,
            [
                {"name": "A", "version": "1.0.0", "depends": [], "extras": {}},
                {"name": "A", "version": "1.0.0", "depends": [], "extras": {}},
            ],
        )

    assert "double specification" in str(e)


@pytest.mark.parametrize(
    ("message", "item"),
    [
        (
            "invalid version",
            {"name": "A", "version": "foo", "depends": [], "extras": {}},
        ),
        (
            "depends has invalid requirement",
            {"name": "A", "version": "1.0.0", "depends": ["..."], "extras": {}},
        ),
        (
            "extra='tada' has invalid requirement",
            {
                "name": "A",
                "version": "1.0.0",
                "depends": [],
                "extras": {"tada": ["..."]},
            },
        ),
    ],
)
def test_index_item_errors(item, message):
    with pytest.raises(YAMLException) as e:
        translate._translate_index_item(0, 0, None, item)

    assert message in str(e)



"""Contains a some helper implementations of zazo.abc classes
"""

from copy import deepcopy

from packaging.requirements import Requirement

from zazo.api import Candidate, Provider


class YAMLCandidate(Candidate):

    def __init__(self, name, version):
        super(YAMLCandidate, self).__init__()
        self.name = name
        self.version = version

        self.extras = set()

    def __repr__(self):
        return "<YAMLCandidate({}[{}], {!r})>".format(
            self.name, ",".join(self.extras), self.version
        )

    def __eq__(self, other):
        return (
            self.name == other.name
            and self.version == other.version
            and self.extras == other.extras
        )

    def matches(self, requirement):
        return (
            self.name == requirement.name
            and self.extras == requirement.extras
            and self.version in requirement.specifier
        )


class YAMLProvider(Provider):

    def __init__(self, candidates, dependencies, installed=None):
        super(YAMLProvider, self).__init__()
        self._candidates_by_name = candidates
        self._dependencies_by_candidate = dependencies

        if installed is None:
            installed = {}

        self._installed = installed  # type: Dict[str, Tuple[Version, Extras]]

    def get_candidates(self, requirement):
        name = requirement.name

        # Make a copy of the candidates
        candidates = self._candidates_by_name.get(name, [])
        candidates = deepcopy(candidates)

        # Merge the extras
        for candidate in candidates:
            candidate.extras |= requirement.extras

        # NOTE: This is simply ordering the matching candidates in decreasing
        #       order of version. It only works in a case of installing in an
        #       empty (or equivalent) environment.
        return list(
            reversed(
                sorted(
                    filter(lambda x: x.matches(requirement), candidates),
                    key=lambda x: x.version,
                )
            )
        )

    def get_dependencies(self, candidate):
        # Short hands
        name = candidate.name
        version = candidate.version

        # Simple Requirement
        if not candidate.extras:
            return self._dependencies_by_candidate["{} {}".format(name, version)]
        # Single Extra (depends on parent + extra)
        elif len(candidate.extras) == 1:
            extra = next(iter(candidate.extras))

            # Add the parent package as the first requirement
            retval = [Requirement("{} == {}".format(name, version))]

            # Add the extra dependencies
            key = "{}[{}] {}".format(name, extra, version)
            retval.extend(self._dependencies_by_candidate.get(key, []))

            return retval
        # Multiple Extras (depends on all single extra variants)
        else:
            return [
                Requirement("{}[{}] == {}".format(name, extra_name, version))
                for extra_name in candidate.extras
            ]



"""Tests for the basic ABCs implemented.
"""

import pytest
from packaging.requirements import Requirement
from packaging.version import Version

from tests.lib.yaml.abcs import YAMLCandidate, YAMLProvider


# Shorthands for writing tests
def can(arg, extras=None):
    kwargs = {}
    kwargs["name"], kwargs["version"] = arg.split(" ")
    kwargs["version"] = Version(kwargs["version"])

    retval = YAMLCandidate(**kwargs)
    if extras is not None:
        retval.extras = set(extras)
    return retval


req = Requirement


# packaging's Requirements don't like being compared. Let's compare them by converting
# them into strings and comparing those.
def assert_requirements_match(requirements_1, requirements_2):
    assert all(isinstance(i, Requirement) for i in requirements_1)
    assert all(isinstance(i, Requirement) for i in requirements_2)

    requirements_strings_1 = sorted(str(i) for i in requirements_1)
    requirements_strings_2 = sorted(str(i) for i in requirements_2)

    assert requirements_strings_1 == requirements_strings_2


class TestCandidate(object):

    @pytest.mark.parametrize(
        ("candidate", "requirement", "should_match"),
        [
            ("A 1.0.0", "A", True),
            ("A 1.0.0", "A == 1.0.0", True),
            ("A 1.0.0", "A != 1.0.0", False),
        ],
    )
    def test_matches_correctly(self, candidate, requirement, should_match):
        if should_match:
            assert can(candidate).matches(req(requirement))
        else:
            assert not can(candidate).matches(req(requirement))

    @pytest.mark.parametrize(
        ("item1", "item2", "should_equal"),
        [
            ("A 1.0.0", "A 1.0.0", True),
            ("A 1.0.0", "B 1.0.0", False),  # diff name
            ("A 1.0.0", "A 2.0.0", False),  # diff ver
            ("A 1.0.0", "B 2.0.0", False),  # diff name and ver
        ],
    )
    def test_equality(self, item1, item2, should_equal):
        if should_equal:
            assert can(item1) == can(item2)
        else:
            assert can(item1) != can(item2)


class TestProvider(object):

    @pytest.mark.parametrize(
        ("candidate_info", "requirement", "expected"),
        [
            (
                {"A": [can("A 3.0.0"), can("A 2.0.0"), can("A 1.0.0")]},
                req("A"),
                [can("A 3.0.0"), can("A 2.0.0"), can("A 1.0.0")],
            ),
            (
                {"A": [can("A 3.0.0"), can("A 2.0.0"), can("A 1.0.0")]},
                req("A[extra]"),
                [
                    can("A 3.0.0", extras={"extra"}),
                    can("A 2.0.0", extras={"extra"}),
                    can("A 1.0.0", extras={"extra"}),
                ],
            ),
            (
                {"A": [can("A 1.0.0")], "B": [can("B 1.0.0")]},
                req("B"),
                [can("B 1.0.0")],
            ),
        ],
    )
    def test_gives_correct_candidates(self, candidate_info, requirement, expected):
        provider = YAMLProvider(candidate_info, {})

        assert provider.get_candidates(requirement) == expected

    @pytest.mark.parametrize(
        ("dependency_info", "requirement", "expected"),
        [
            (
                {
                    "A 2.0.0": [req("B")],
                    "A 1.0.0": [req("C")],
                    "A 0.2.0": [req("D")],
                    "A 0.1.0": [req("E")],
                },
                can("A 1.0.0"),
                [req("C")],
            ),
            (
                {
                    "A 2.0.0": [req("one")],
                    "A[doc] 2.0.0": [req("two")],
                    "A 1.0.0": [req("three")],
                    "A[doc] 1.0.0": [req("four")],
                },
                can("A 1.0.0", extras={"doc"}),
                [req("A == 1.0.0"), req("four")],
            ),
            (
                {
                    "A 1.0.0": [req("one")],
                    "A[doc] 1.0.0": [req("two")],
                    "A[test] 1.0.0": [req("three")],
                },
                can("A 1.0.0", extras={"doc", "test"}),
                [req("A[doc] == 1.0.0"), req("A[test] == 1.0.0")],
            ),
        ],
    )
    def test_gives_correct_dependencies(self, dependency_info, requirement, expected):
        provider = YAMLProvider({}, dependency_info)

        assert_requirements_match(provider.get_dependencies(requirement), expected)

    # TODO: Add tests for installation
    #       - double installs fail
    #       - extras on installation are modified correctly
    #       - extras installed are provided in returned candidate
    #       - version mismatch
    # TODO: Add tests for un-installation
    #       - refuses to uninstall stuff that's not installed
    #       - actually uninstalls stuff that's install



"""Helpers for YAML-file based tests
"""

from __future__ import absolute_import

import traceback

import pytest
import yaml
from packaging.requirements import InvalidRequirement, Requirement

from zazo.api import BackTrackingResolver, CannotSatisfy

from .abcs import YAMLProvider
from .exceptions import YAMLException
from .expand import expand_shorthands
from .translate import translate_specification
from .verify import verify_installation_result


class YamlFixtureItem(pytest.Item):

    def __init__(self, name, parent, spec):
        super(YamlFixtureItem, self).__init__(name, parent)
        self.spec = spec

        self._actions = {
            "install": (self._do_install, self._check_install),
            "uninstall": (self._do_uninstall, self._check_uninstall),
        }

    # Called when the tests fail
    def repr_failure(self, excinfo):
        error = excinfo.value
        if isinstance(error, YAMLException):
            return str(error)
        elif isinstance(error, AssertionError):
            msg = ": {}".format(error.args[0]) if error.args else ""
            return "assertion failed" + msg
        else:
            trace = traceback.format_exception(excinfo.type, excinfo.value, excinfo.tb)
            return "".join(trace) + "\nError occurred."

    def reportinfo(self):
        return self.fspath, 0, "yaml-test: %s" % self.name

    def runtest(self):
        # Convert things into something usable.
        expanded_spec = expand_shorthands(self.name, self.spec)
        test_spec = translate_specification(self.name, expanded_spec)

        # Create a Provider, using information
        provider = YAMLProvider(test_spec["candidates"], test_spec["dependencies"])

        for action, expected in test_spec["subtests"]:
            result = self._perform_action(action, provider)
            self._check_action_result(action, result, expected)

    ###### Helpers ######
    def _perform_action(self, action, provider):
        verb = action["action"]
        func = getattr(self, "_do_" + verb, None)
        if func is None:
            raise Exception("Can not perform verb {!r}.".format(verb))

        return func(provider, action)

    def _check_action_result(self, action, result, expected):
        verb = action["action"]
        func = getattr(self, "_check_" + verb, None)
        if func is None:
            raise Exception("Can not check verb {!r}.".format(verb))

        func(result, expected)

    ###### Actions ######
    def _do_uninstall(self, provider, names):
        # _check_list(names, "uninstall items", str)
        provider.uninstall_names(names)

    def _check_uninstall(self, result, expected):
        assert result is None
        assert expected == "tada"

    def _do_install_prepare_reqs(self, action):
        """Create a list of requirements from given items.
        """

        def _make_req(string):
            try:
                return Requirement(string)
            except Exception:
                raise YAMLException(
                    "Could not parse requirement {!r} from {!r}",
                    string,
                    action["items"],
                )

        return [_make_req(r) for r in action["items"]]

    def _do_install(self, provider, items):
        """Compose a list of requirements and then run the resolver to install them.
        """
        requirements = self._do_install_prepare_reqs(items)
        result = self._run_resolver(requirements, provider)

        return result

    def _check_install(self, result, expected):
        verify_installation_result(result, expected)

    def _run_resolver(self, requirements, provider):
        resolver = BackTrackingResolver(provider)
        try:
            to_install = resolver.resolve(requirements)
        except CannotSatisfy as e:
            return e
        else:
            return to_install


class YamlFixtureFile(pytest.File):

    def _compose_tests(self, data):
        # _check_list(data, "test data")
        for i, item in enumerate(data):
            name = "{}[{}]".format(self.fspath.basename, i)
            yield YamlFixtureItem(name, self, item)

    def collect(self):
        loader = yaml.CLoader if hasattr(yaml, "CLoader") else yaml.Loader
        with self.fspath.open() as f:
            data = yaml.load(f.read(), Loader=loader)
        for test in self._compose_tests(data):
            yield test



"""Helpers for converting between forms.

YAML -> (Candidates, Requirements) -> Error Messages
"""

from collections import defaultdict

from packaging.requirements import Requirement
from packaging.version import Version

from zazo.api import CannotSatisfy

from .exceptions import YAMLException


def _convert_error(result):
    return {"conflicts": []}


def _convert_resolved_set(result):
    retval = result.copy()
    for item in result:
        if "[" in item:
            del retval[item]
    # TODO: Make this do some actual restructuring
    return {"chosen_set": retval}


def verify_installation_result(result, expected):
    """Verify that an installation run did what it was supposed to.
    """

    if not isinstance(expected, dict):
        raise YAMLException("The expected result of this test is not a dictionary.")

    if isinstance(result, CannotSatisfy):
        result = _convert_error(result)
    else:
        result = _convert_resolved_set(result)

    errors = []
    if "conflicts" in expected:
        if "conflicts" not in result:
            message = "Expected to get conflicts, got resolved set"
            raise AssertionError(message)
        # TODO: Beef this up; maybe try to show what's messed up.
    else:
        assert "set" in expected, "set not in expected"
        if "chosen_set" not in result:
            message = "Expected to get resolved set, got conflicts"
            raise AssertionError(message)

        # Make sure we got the right versions
        for item in expected["set"]:
            name, version = item.split(" ", 1)
            if name not in result["chosen_set"]:
                errors.append(name + " is missing.")
                continue
            if Version(version) != result["chosen_set"][name].version:
                errors.append(
                    "Expected {} to be version {}, got {}".format(
                        name, version, result["chosen_set"][name].version
                    )
                )
            del result["chosen_set"][name]

        # Make sure we got the right packages
        if result["chosen_set"]:
            for key in result["chosen_set"]:
                errors.append(
                    "Got unexpected selection: {} {}".format(
                        key, result["chosen_set"][key].version
                    )
                )

        # TODO: Check the graph of dependencies

    if errors:
        raise IncorrectResolution(errors)



"""Test various expansions
"""

import pytest

from tests.lib.yaml import expand
from tests.lib.yaml.exceptions import MalformedYAML


@pytest.mark.parametrize(
    "di",
    [
        {},
        {"index": None},
        {"actions": None},
        {"results": None},
        {"index": None, "actions": None},
        {"index": None, "results": None},
        {"actions": None, "results": None},
        {"index": None, "actions": None, "results": None, "hello": None},
    ],
)
def test_fails_when_top_level_structure_is_incorrect(di):
    with pytest.raises(MalformedYAML) as err:
        expand.expand_shorthands(0, di)

    assert "top-level has incorrect structure" in str(err)


@pytest.mark.parametrize(
    ("string", "expected"),
    [
        ("A 1.0.0", {"name": "A", "version": "1.0.0", "depends": [], "extras": {}}),
        (
            "A 1.0.0 depends B == 1.0.0 & C",
            {
                "name": "A",
                "version": "1.0.0",
                "depends": ["B == 1.0.0", "C"],
                "extras": {},
            },
        ),
    ],
)
def test_expand_index_string(string, expected):
    assert expand._expand_index_item_string(0, 0, string) == expected


@pytest.mark.parametrize(
    ("string", "expected_words"),
    [
        ("A", "does not specify name and version correctly"),
        ("A depends B", "does not specify name and version correctly"),
        ("A 1.0.0 depends", "does not specify 'depends' correctly"),
    ],
)
def test_expand_index_string_errors(string, expected_words):
    with pytest.raises(MalformedYAML) as err:
        expand._expand_index_item_string(0, 0, string)

    assert expected_words in str(err)


@pytest.mark.parametrize(
    "item",
    [
        # Wrong Keys
        {"name": "A", "version": "1.0.0"},
        {"name": "A", "version": "1.0.0", "depends": []},
        {"name": "A", "version": "1.0.0", "depends": [], "extras": []},
        # Incorrect types
        {"name": "A", "version": "1.0.0", "depends": [], "extras": []},
        {"name": "A", "version": "1.0.0", "depends": {}, "extras": {}},
        {"name": ["A"], "version": "1.0.0", "depends": [], "extras": {}},
        {"name": "A", "version": ["1.0.0"], "depends": [], "extras": {}},
        {"name": "A", "version": "1.0.0", "depends": [1], "extras": {}},
        {"name": "A", "version": "1.0.0", "depends": [1], "extras": {}},
        {"name": "A", "version": "1.0.0", "depends": [], "extras": {1: []}},
        {"name": "A", "version": "1.0.0", "depends": [], "extras": {"str": [1]}},
    ],
)
def test_expand_index_item_errors(item):
    with pytest.raises(MalformedYAML) as err:
        expand._expand_index_item(0, 0, item)

    assert "index[0] has incorrect structure" in str(err)


@pytest.mark.parametrize(
    ("di", "expected"),
    [
        ({"install": [1, 2, 3]}, {"action": "install", "items": [1, 2, 3]}),
        ({"action": "abcd", "items": "efgh"}, {"action": "abcd", "items": "efgh"}),
    ],
)
def test_expand_action_dict(di, expected):
    assert expand._expand_action_item_dict(0, 0, di) == expected


@pytest.mark.parametrize(
    ("di", "expected_words"),
    [
        ({}, "empty dictionary"),
        ({"action": "abcd"}, "has an 'action' but no 'items'"),
        ({"blah": "abcd", "foo": "efgh"}, "incorrect dictionary"),
    ],
)
def test_expand_action_dict_errors(di, expected_words):
    with pytest.raises(MalformedYAML) as err:
        expand._expand_action_item_dict(0, 0, di)

    assert expected_words in str(err)


@pytest.mark.parametrize(
    ("di", "expected"),
    [
        ({"foo": "bar"}, {"foo": "bar"}),
        ({"set": "abcd", "graph": "efgh"}, {"set": "abcd", "graph": "efgh"}),
    ],
)
def test_expand_result_dict(di, expected):
    assert expand._expand_result_item_dict(0, 0, di) == expected


@pytest.mark.parametrize(
    ("di", "expected_words"),
    [
        ({}, "empty dictionary"),
        ({"set": []}, "has a 'set' but no 'graph'"),
        ({"graph": []}, "has a 'graph' but no 'set'"),
    ],
)
def test_expand_result_dict_errors(di, expected_words):
    with pytest.raises(MalformedYAML) as err:
        expand._expand_result_item_dict(0, 0, di)

    assert expected_words in str(err)


def test_expand_valid():
    pass



import functools
import glob
import subprocess

import nox

LINT_ITEMS = "nox.py", "docs/source/conf.py", "zazo", "tests"


@nox.session
def docs(session):
    session.install("-r", "tools/reqs/docs.txt")

    session.run("sphinx-build", "-n", "-b", "html", "docs/source", "docs/build")


@nox.session
def packaging(session):
    session.install("-r", "tools/reqs/packaging.txt")

    session.run("flit", "build")


def lint_session(func):

    @functools.wraps(func)
    def wrapped(session):
        if session.posargs:
            files = session.posargs
        else:
            files = LINT_ITEMS
        session.install("--pre", "-r", "tools/reqs/lint.txt")

        session.run("black", "--version")
        session.run("isort", "--version")
        session.run("mypy", "--version")

        return func(session, files)

    return wrapped


@nox.session
@lint_session
def lint(session, files):
    session.run("black", "--check", "--diff", *files)
    session.run("isort", "--check-only", "--diff", "--recursive", *files)
    session.run("mypy", "--ignore-missing-imports", "--check-untyped-defs", "zazo")
    session.run(
        "mypy", "-2", "--ignore-missing-imports", "--check-untyped-defs", "zazo"
    )


@nox.session
@lint_session
def format(session, files):
    session.run("black", *files)
    session.run("isort", "--recursive", *files)


@nox.session
@nox.parametrize("python_version", ["2.7", "3.4", "3.5", "3.6", "3.7", "pypy", "pypy3"])
def test(session, python_version):
    # Set the interpreter
    if python_version.startswith("pypy"):
        session.interpreter = python_version
    else:
        session.interpreter = "python" + python_version

    # Build the package.
    # THIS IS A HACK
    #   Working around all kinds of weird nox + flit + Travis CI behavior.
    #   We're building a wheel here and installing it with session.install since
    #   nox is declarative but we need to run the build command before executing
    #   code.
    def my_run(*args):
        print("run > " + " ".join(args))
        try:
            subprocess.check_call(args)
        except subprocess.CalledProcessError:
            session.error("Command failed.")

    my_run("python3", "-m", "flit", "build")
    files = glob.glob("./dist/*.whl")
    if not files:
        session.error("Could not find any built wheels.")

    # Install the package and test dependencies.
    session.install(*files)
    session.install("-r", "tools/reqs/test.txt")

    # Run the tests
    session.cd("tests")  # we change directory to avoid the cwd ambiguity
    session.run("pytest", *session.posargs)



"""Expansion logic for shorthands in YAML.
"""

from .exceptions import AggregatedYAMLErrors, MalformedYAML, YAMLException


def expand_shorthands(test_id, test_dict):
    """Expands all the shorthands to the "expanded" form.

    This simplifies the rest of the code by eliminating the need for them to handle
    shorthands and basic sanity checks.
    """

    _ensure_keys(
        test_id, "top-level", test_dict, {"name", "actions", "index", "results"}
    )

    # Expand everything.
    index = _expand_index(test_id, test_dict["index"])
    actions = _expand_actions(test_id, test_dict["actions"])
    results = _expand_results(test_id, test_dict["results"])

    return {"index": index, "actions": actions, "results": results}


def _ensure_keys(test_id, level, di, expected_keys):
    actual_keys = set(di.keys())
    if actual_keys != expected_keys:
        missing = sorted(expected_keys - actual_keys)
        extra = sorted(actual_keys - expected_keys)

        parts = []
        if missing:
            parts.append("(have missing: " + str(missing)[1:-1] + ")")
        if extra:
            parts.append("(got extra: " + str(extra)[1:-1] + ")")

        raise MalformedYAML(
            test_id, "{} has incorrect structure: {}", level, " ".join(parts)
        )


def _meta_expand(test_id, argument, name, function):
    if not isinstance(argument, list):
        raise MalformedYAML(test_id, "has a non-list " + name)

    errors = AggregatedYAMLErrors()
    for i, item in enumerate(argument):
        try:
            item = function(test_id, i, item)
        except YAMLException as e:
            errors.add(e)
        else:
            # Substitute with the changed item.
            argument[i] = item

    if errors:
        raise errors

    return argument


def _expand_index(test_id, index):
    return _meta_expand(test_id, index, "index", _expand_index_item)


def _expand_index_item(test_id, i, item):
    if isinstance(item, str):
        item = _expand_index_item_string(test_id, i, item)
    elif isinstance(item, dict):
        pass
    else:
        raise MalformedYAML(test_id, "index[{}] is incorrect: {!r}", i, item)

    _ensure_keys(
        test_id, "index[{}]".format(i), item, {"name", "version", "depends", "extras"}
    )
    if not (
        isinstance(item["name"], str)
        and isinstance(item["version"], str)
        and isinstance(item["depends"], list)
        and all(isinstance(obj, str) for obj in item["depends"])
        and isinstance(item["extras"], dict)
        and all(isinstance(obj, str) for obj in item["extras"].keys())
        and all(isinstance(obj, list) for obj in item["extras"].values())
        and all(
            isinstance(elem, str) for obj in item["extras"].values() for elem in obj
        )
    ):
        raise MalformedYAML(test_id, "index[{}] has incorrect structure.", i)

    return item


def _expand_index_item_string(test_id, i, item):
    """Expand an item in the index, which is a string.
    """

    def _split_and_strip(my_str, splitwith, count=None):
        """Split a string and strip each component
        """
        if count is None:
            return [x.strip() for x in my_str.strip().split(splitwith)]
        else:
            return [x.strip() for x in my_str.strip().split(splitwith, count)]

    # Super Cautious Checking of depends.
    if "depends" in item:  # this could be wrong but eh. Will deal with that later.
        errored = False
        try:
            name_version, depends_str = _split_and_strip(item, " depends ", 1)
            depends_str = _split_and_strip(depends_str, "&")
        except ValueError:
            errored = True

        if errored:
            raise MalformedYAML(
                test_id, "index[{}] does not specify 'depends' correctly", i
            )
    else:
        name_version = item
        depends_str = []

    try:
        name, version = _split_and_strip(name_version, " ", 1)
    except ValueError:
        raise MalformedYAML(
            test_id, "index[{}] does not specify name and version correctly", i
        )

    return {"name": name, "version": version, "depends": depends_str, "extras": {}}


def _expand_actions(test_id, actions):
    return _meta_expand(test_id, actions, "actions", _expand_action_item)


def _expand_action_item(test_id, i, item):
    if isinstance(item, dict):
        return _expand_action_item_dict(test_id, i, item)
    else:
        raise MalformedYAML(test_id, "actions[{}] is not a dictionary: {!r}", i, item)


def _expand_action_item_dict(test_id, i, item):
    if not item:
        raise MalformedYAML(test_id, "actions[{}] is an empty dictionary.", i)
    elif "action" in item:
        if "items" not in item:
            raise MalformedYAML(
                test_id, "actions[{}] has an 'action' but no 'items'.", i
            )
        return item  # We'll validate later.
    elif len(item) == 1:
        # convert {"install": ...} -> {"action": "install", "items": ...}
        key, value = next(iter(item.items()))
        return {"action": key, "items": value}
    else:
        raise MalformedYAML(test_id, "actions[{}] is an incorrect dictionary.", i)


def _expand_results(test_id, results):
    return _meta_expand(test_id, results, "results", _expand_result_item)


def _expand_result_item(test_id, i, item):
    if isinstance(item, dict):
        return _expand_result_item_dict(test_id, i, item)
    else:
        raise MalformedYAML(test_id, "results[{}] is not a dictionary.", i)


def _expand_result_item_dict(test_id, i, item):
    if not item:
        raise MalformedYAML(test_id, "results[{}] is an empty dictionary.", i)

    if "set" in item and "graph" not in item:
        raise MalformedYAML(test_id, "results[{}] has a 'set' but no 'graph'.", i)
    elif "graph" in item and "set" not in item:
        raise MalformedYAML(test_id, "results[{}] has a 'graph' but no 'set'.", i)

    return item



"""
"""

import inspect
import logging

from .exceptions import CannotSatisfy

if False:
    from typing import Any, List, Dict  # noqa
    from packaging.requirements import Requirement  # noqa
    from zazo.abc import Provider, Candidate  # noqa

    Graph = Dict[str, Candidate]

logger = logging.getLogger()


def _create_key(req):
    parts = [req.name]

    if req.extras:
        parts.append("[{0}]".format(",".join(sorted(req.extras))))

    return "".join(parts)


class BackTrackingResolver(object):

    def __init__(self, provider):
        # type: (Provider) -> None
        super(BackTrackingResolver, self).__init__()
        self.provider = provider

    def resolve(self, requirements):
        # type: (List[Requirement]) -> Graph
        self._initial_recursion_depth = len(inspect.stack())
        graph = {}  # type: Graph
        try:
            # The following involves a speedup hack.
            retval = self._resolve(requirements, graph)
        except CannotSatisfy:
            logger.debug("Unable to satisfy dependencies.")
            raise
        else:
            logger.debug("Found satisfying set: %s", retval)
            return retval

    # TODO: Figure out feasibility of blacklisting of known conflicts, like
    #       Stork.
    def _resolve(self, requirements, graph):
        # type: (List[Requirement], Graph, Any) -> Graph
        if not requirements:
            return graph

        # Extract the top level requirement
        req = requirements[0]
        req_key = _create_key(req)
        requirements = requirements[1:]

        if req_key in graph:
            # We have seen this requirement; check if we satisfy it already
            chosen_before = graph[req_key]
            if not chosen_before.matches(req):
                raise CannotSatisfy(req, chosen_before)
            # Proceed to the next requirement
            try:
            except CannotSatisfy:
                raise
            else:
                return retval

        candidates = self.provider.get_candidates(req)
        for candidate in candidates:
            assert candidate.matches(req), (
                "candidate does not match requirement it was guaranteed " "to match"
            )

            graph[req_key] = candidate
            deps = self.provider.get_dependencies(candidate)
            try:
                # XXX: This causes a peak in memory usage.
            except CannotSatisfy:
                assert graph[req_key] == candidate
                del graph[req_key]
            else:
                return retval

        # If we are here, we could not satisfy the given requirements
        raise CannotSatisfy([req] + requirements, graph)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['zazo']

package_data = \
{'': ['*']}

install_requires = \
['packaging', 'six']

setup(name='zazo',
      version='0.0.0a2',
      description='An Extensible Dependency Resolver written in Python',
      author='Pradyun Gedam',
      author_email='pradyunsg@gmail.com',
      url='https://github.com/pradyunsg/zazo',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )



class YAMLException(Exception):
    """Raised when a YAML test is not formatted correctly.
    """

    def __str__(self):
        # Format a reason
        if not self.args:
            reason = "unknown"
        elif len(self.args) == 1:
            reason = self.args[0]
        else:
            try:
                reason = self.args[0].format(*self.args[1:])
            except Exception as error:
                reason = "Unable to format message: {}\n{}".format(self.args, error)

        # Append the reason
        return "YAML is malformed -- reason: {}".format(reason)


class MalformedYAML(YAMLException):
    """Raised when a malformed YAML specification is found.
    """

    def __init__(self, test_id, message, *args, **kwargs):
        super(MalformedYAML, self).__init__()
        self.id = test_id
        self.message = message.format(*args, **kwargs)

    def __str__(self):
        return "Malformed YAML in {}: {}".format(self.id, self.message)


class AggregatedYAMLErrors(YAMLException):
    """Raised when there are multiple issues with the YAML specification.
    """

    def __init__(self):
        super(AggregatedYAMLErrors, self).__init__()
        self._errors = []

    def __bool__(self):
        return bool(self._errors)

    __nonzero__ = __bool__  # Python 2 stuff

    def add(self, error):
        if isinstance(error, AggregatedYAMLErrors):
            self._errors.extend(error._errors)
        else:
            self._errors.append(error)

    def __str__(self):
        if not self._errors:
            return "<no errors>"
        elif len(self._errors) == 1:
            return str(self._errors[0])
        else:
            parts = ["There are multiple issues:"]
            for err in self._errors:
                parts.append(str(err))
            return "\n- ".join(parts)



"""
"""
from __future__ import absolute_import

import abc

from six import add_metaclass

if False:
    from packaging.requirements import Requirement  # NOQA
    from packaging.version import Version  # NOQA
    from typing import List  # NOQA


@add_metaclass(abc.ABCMeta)
class Candidate(object):

    @abc.abstractmethod
    def matches(self, requirement):
        # type: (Requirement) -> bool
        """Does this candidate match the given requirement?
        """
        raise NotImplementedError("Method to be overridden in a subclass.")


@add_metaclass(abc.ABCMeta)
class Provider(object):
    """Handles everything related to providing packages and package information
    """

    @abc.abstractmethod
    def get_candidates(self, requirement):
        # type: (Requirement) -> List[Candidate]
        """Get all the candidates that match the given requirement.
        """
        raise NotImplementedError("Method to be overridden in a subclass.")

    @abc.abstractmethod
    def get_dependencies(self, candidate):
        # type: (Candidate) -> List[Requirement]
        """Fetch dependencies of the given candidate
        """
        raise NotImplementedError("Method to be overridden in a subclass.")
